# -*- coding: iso-8859-15 -*-

## Liste des oprateurs:
operators = []

## Description de la situation initiale:
initialState = []

## Liste des tats encore  explorer:
openList = []

## Heuristique d'ordonnancement pour A*:
heuristicFn = []

## Fonction vrifiant la consistence d'un noeud, cd si ses buts sont
## cohrents:
consistentFn = []

## Liste des solutions trouves:
plannerSolutions = []

###===================================================================
###   OPERATEURS
###===================================================================

class Operator:

	## Constructeur
	## @param action Action effectue
	## @param preconds Liste des prconditions
	## @param addList Liste ajoute
	## @param deleteList liste enleve
	## @cost cost Cot de l'oprateur.
	def __init__(self, action, preconds, addList, deleteList, cost):
		self.action=action
		self.preconds=preconds
		self.addList=addList
		self.deleteList=deleteList
		self.cost=cost

	## Retourne une reprsentation en string des caractristiques de l'oprateur
	def __str__(self):
		out = ("<Operator %s; Preconds=%s; Add-List=%s; Delete-List=%s; cost=%s>\n" % (self.action, self.preconds, self.addList, self.deleteList, self.cost))
		return out

## Dfinit un nouvel oprateur d'aprs sa <description> et l'ajoute
## a la variable globale operators.
## @param description Liste [action preconditions add-list delete-list cot]
def defOperator(description):
	global operators
	op = Operator(description[0],description[1],description[2],description[3],description[4])
	operators.append(op)


###===================================================================
###   ETATS (STATES)
###===================================================================

class State:

	## Constructeur
	## @param goals Les buts de l'tat.
	## @param operators La liste des oprateurs appliqus  cet tat pour parvenir a l'tat final
	## @param cost Le cot de cet tat
	def __init__(self, goals, ops, cost=0):
		self.goals = goals
		self.operators = ops
		self.cost = cost

	## Compare deux tats
	## @param other L'autre tat  comparer
	def __cmp__(self, other):
		return self.cost - other.cost

	## Retourne une reprsentation en string avec les caractristiques de l'tat
	def __str__(self):
		out = ("<State: Goals=%s; operators=%s; cost=%s>\n" % (self.goals, self.operators, self.cost))
		return out

	## Retourne True ssi l'tat est une solution, cd si ses buts
	## font partie de l'tat initial.
	def isSolution(self):
		global initialState
		return set(self.goals).issubset(set(initialState))

	## Calcule le cot de l'tat d'aprs la fonction heuristique
	## heuristicFn et retourne l'tat.
	def computeCost(self):
		global heuristicFn
		self.cost = apply(heuristicFn, [self])
		return self

	## Retourne tous les nouveaux tats que l'on peut obtenir en
	## appliquant les oprateurs  l'tat .
	## @param ops Les oprateurs  appliquer
	def applyOperators(self, ops):
		rs = []
		for o in ops:
			s = self.applyOperator(o)
			if isinstance(s, State):
				rs.append(s)
		return rs

	## Applique l'oprateur <op> a l'tat s'il est applicable et
	## retourne le nouvel tat.
	## @param op L'oprateur  appliquer
	def applyOperator(self, op):
		global consistentFn
		x = self.applicable(op)
		if x:
			newGoals = list(set(self.goals).union(set(op.preconds)).difference(set(x)))
			newOperators = self.operators[:]
			newOperators.append(op)
			newState = State(newGoals, newOperators)
			newState.computeCost()
			if apply(consistentFn, [newState]):
				return newState
		return []
	
	## Verifie que l'oprateur op peut s'appliquer a l'tat, cd que:
	## - aucun prdicat de la deleteList de op ne fait partie des
	##	 buts de l'tat.
	## - au moins un but de l'tat est satisfait par un prdicat de
	##	 l'addList de op.
	## Retourne la liste des buts satisfaits par op.
	## @param op L'oprateur  appliquer
	def applicable(self,op):
		if len(set(self.goals).intersection(set(op.deleteList))) < 1:
			return set(self.goals).intersection(set(op.addList))
		else:
			return []

	## Analyse moyens-buts; retourne le nombre de buts de l'tat 
	## qui ne sont pas satisfaits par l'tat initial initialState.
	def meansEnds(self):
		global initialState
		return len(set(initialState).difference(set(self.goals)))

	## Retourne la somme des cots des oprateurs appliqus a l'tat
	## final pour obtenir l'tat.
	def opCost(self):
		if self.operators:
			return reduce(lambda x,y: x + y.cost, self.operators, self.cost)
		else:
			return self.cost
	
	## Affiche la solution
	def printPlannerSolution(self):
		print("== PLANNER SOLUTION ==")
		i = 0
		for op in self.operators:
			print("Action %s: %s" % (i, op.action))
			i = i+1
  


###===================================================================
###   INITIALISATION et CONSTRUCTION
###===================================================================

## Initialise et construit un nouveau planificateur.
## @param initState L'tat initial	  
## @param goals Les buts
## @param ops La description des oprateurs
## @param consistent La fonction de consistance
## @param heuristic La fonction heuristique utilise par A*
def buildPlanner(initState, goals, ops, consistent, heuristic):
	global operators		
	global initialState
	global plannerSolutions
	global openList
	global consistentFn
	global heuristicFn
	operators = []			
	initialState = initState
	plannerSolutions = []
	openList = [State(goals,[],0)]
	consistentFn = consistent
	heuristicFn = heuristic
	for op in ops:
		defOperator(op)
	return openList[0]


###===================================================================
###   RECHERCHE
###===================================================================

## Recherche une solution en explorant au plus <max-steps> tats.
## @param maxSteps Le nombre maximal de pas
def findPlan(maxSteps=1000):
	global plannerSolutions
	sol, steps = AStarSearch(maxSteps)
	if isinstance(sol,State):
		print("FOUND A SOLUTION IN %s STEPS\n" % (steps))
		plannerSolutions = [sol] + plannerSolutions
		sol.printPlannerSolution()
		return sol
	elif steps < 0:
		print("NO MORE SOLUTIONS (open list exhausted)\n")
	else:
		print("NO SOLUTION FOUND IN %s STEPS\n" % (steps))

## Algorithme de recherche A*. Effectue au maximum <max-steps> pas
## de recherche. N'utilise pas de closed-list.
## Retourne 2 valeurs:
## - la solution trouve ou [] si aucune solution n'a t
##	 trouve.
## - le nombre de pas de recherche effectus. Si openList a
##	 t "vide" (cd il n'y a plus de solutions), cette
##	 deuxieme valeur sera -1.
## @param maxSteps Le nombre maximal de pas
def AStarSearch(maxsteps):
	global openList
	step = 0
	while len(openList) > 0:
		s = openList[0]
		openList = openList[1:]
		if s.isSolution():
			return s, step
		openList = openList + s.applyOperators(operators)
		openList.sort()
		step = step + 1
		if step == maxsteps:
			return [], step
	return [], -1


###===================================================================
###   UTILITAIRES
###===================================================================

## Affiche toutes les solutions trouves par le planificateur
def printPlannerSolutions():
	global plannerSolutions
	print(">>> FOUND %s PLANS <<<" % (len(plannerSolutions)))
	for sol in plannerSolutions.revers():
		sol.printPlannerSolution()
